@@ -3,6 +3,7 @@
 use std::{fmt::Write, iter::once, mem::take};
 
 use retain_mut::RetainMut;
+use rustc_hash::FxHashMap;
 use swc_atoms::{js_word, JsWord};
 use swc_common::{
     collections::AHashMap, iter::IdentifyLast, pass::Repeated, util::take::Take, Spanned,
@@ -81,6 +82,7 @@ where
         label: Default::default(),
         mode,
         debug_infinite_loop,
+        functions: Default::default(),
     }
 }
 
@@ -215,6 +217,8 @@ struct Optimizer<'a, M> {
     mode: &'a M,
 
     debug_infinite_loop: bool,
+
+    functions: FxHashMap<Id, FnMetadata>,
 }
 
 impl<M> Repeated for Optimizer<'_, M> {
@@ -227,6 +231,23 @@ impl<M> Repeated for Optimizer<'_, M> {
     }
 }
 
+#[derive(Debug, Clone, Copy)]
+struct FnMetadata {
+    len: usize,
+}
+
+impl From<&Function> for FnMetadata {
+    fn from(f: &Function) -> Self {
+        FnMetadata {
+            len: f
+                .params
+                .iter()
+                .filter(|p| matches!(&p.pat, Pat::Ident(..) | Pat::Array(..) | Pat::Object(..)))
+                .count(),
+        }
+    }
+}
+
 impl<M> Optimizer<'_, M>
 where
     M: Mode,
@@ -1954,6 +1975,10 @@ where
     }
 
     fn visit_mut_fn_decl(&mut self, f: &mut FnDecl) {
+        self.functions
+            .entry(f.ident.to_id())
+            .or_insert_with(|| FnMetadata::from(&f.function));
+
         if !self.options.keep_fargs && self.options.evaluate && self.options.unused {
             self.drop_unused_params(&mut f.function.params);
         }
@@ -1963,6 +1988,12 @@ where
     }
 
     fn visit_mut_fn_expr(&mut self, e: &mut FnExpr) {
+        if let Some(ident) = &e.ident {
+            self.functions
+                .entry(ident.to_id())
+                .or_insert_with(|| FnMetadata::from(&e.function));
+        }
+
         if !self.options.keep_fnames {
             self.remove_name_if_not_used(&mut e.ident);
         }
